/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mnenhy.
 *
 * The Initial Developer of the Original Code is Karsten Düsterloh <kd@mnenhy.de>.
 *
 * Portions created by the Initial Developer are Copyright © 2002-2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Markus Hossner <markushossner@gmx.de> (author of the original messageidfinder extension)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

Components.utils.import("resource://mnenhy/mnenhy.js");

// +-
// |  These functions are overloaded here:
// |    mailContextMenus.js:FillMessageIdContextMenu(messageIdNode)
// |    mailContextMenus.js:OpenMessageForMessageId(messageId)
// |    mailContextMenus.js:OpenBrowserWithMessageId(messageId)
// |  Last checked for changes: 2010-01-05
// +-

// return a window with a msgWindow, if possible
function CheckThisMsgWindow()
{
  // we need a msgWindow here!
  let oThis = this;
  if (!("msgWindow" in oThis))
  {
    // we may be in viewSource.xul, thus try the opening window
    // (viewSource without an opener is unlikely)
    oThis = oThis.opener;
    if (!("msgWindow" in oThis))
      return null;
  }
  return oThis;
}


function HideHeaderViews()
{
  if (goMnenhy.installed.IsThunderbird())
  {
    document.getElementById('msgHeaderView').collapsed = true;
  }
  else
  {
    let oCollapsedHeaderViewNode = window.document.getElementById("collapsedHeaderView");
    let oExpandedHeaderViewNode  = window.document.getElementById("expandedHeaderView");
    if (oCollapsedHeaderViewNode)
      oCollapsedHeaderViewNode.collapsed = true;
    if (oExpandedHeaderViewNode)
      oExpandedHeaderViewNode.collapsed = true;
  }
}


// we need to overlay this function to be able to load the browser URL in the messagepane
function OpenBrowserWithMessageId(asMessageID)
{
  let sBrowserURL = goMnenhy.GetPref("mailnews.messageid_browser.url", goMnenhy.kiPrefWString);
  sBrowserURL = sBrowserURL.replace(/%mid/, asMessageID);
  if (!sBrowserURL)
    return;

  if (goMnenhy.GetPref("extensions.mnenhy.mids.loadInMessagePane", goMnenhy.kiPrefBool))
  {
    HideHeaderViews();
    window.content.location = sBrowserURL;
  }
  else
  {
    // we need a msgWindow here!
    let oThis = CheckThisMsgWindow();
    if (!oThis)
      return;

    // SM opens browser URLs himself
    if ("openAsExternal" in oThis)
    {
      oThis.openAsExternal(sBrowserURL);
      return;
    }

    // TB3 passes URLs to an external program
    try
    {
      oThis.messenger.launchExternalURL(sBrowserURL);
    }
    catch(e)
    {
      Components.utils.reportError("Failed to open message-id in browser; "
                                 + "sBrowserURL=" + sBrowserURL);
    }
  }
}


function FindMsgHdrByMID(asMessageId)
{
  // we need a msgWindow here!
  let oThis = CheckThisMsgWindow();
  if (!oThis)
    return null;

  window.setCursor("wait");

  // first search in current folder for message id
  let messageHeader = null;
  // If we have a message at hand, take its folder. If not, try the global folder.
  // If not even that, skip optimization.
  let startFolder = ("gDBView" in oThis) && oThis.gDBView && oThis.gDBView.msgFolder || oThis.msgWindow.openFolder;
  if (startFolder)
  {
    try
    {
      messageHeader = oThis.CheckForMessageIdInFolder(startFolder, asMessageId);
    }
    catch (ex)
    {
      // ignore OOM etc.
      let sInfo = startFolder && startFolder.prettyName || "";
      if (sInfo)
        sInfo = ' when searching in group ' + sInfo;
      Components.utils.reportError(ex + sInfo);
      messageHeader = null;
    }
    // leave current folder in opened state
  }

  // if message id not found in current folder search in all folders
  if (!messageHeader)
  {
    let accountManager = Components.classes["@mozilla.org/messenger/account-manager;1"]
                                   .getService(Components.interfaces.nsIMsgAccountManager);
    let allServers  = oThis.accountManager.allServers;
    let startServer = startFolder && startFolder.server;
    // we may not always find a start folder, e.g. in the standalone message window
    if (startServer)
    {
      try
      {
        messageHeader = oThis.SearchForMessageIdInSubFolder(startServer.rootFolder, asMessageId);
      }
      catch (ex)
      {
        // ignore OOM etc.
        let sInfo = startServer && startServer.prettyName || "";
        if (sInfo)
          sInfo = ' when searching on server ' + sInfo;
        Components.utils.reportError(ex + sInfo);
        startServer   = null;
        messageHeader = null;
      }
    }
    for (let i = 0; i < allServers.Count() && !messageHeader; ++i)
    {
      let currentServer = allServers.GetElementAt(i);
      if ((currentServer instanceof Components.interfaces.nsIMsgIncomingServer) &&
          startServer != currentServer && currentServer.canSearchMessages &&
          !currentServer.isDeferredTo)
      {
        try
        {
          messageHeader = oThis.SearchForMessageIdInSubFolder(currentServer.rootFolder, asMessageId);
        }
        catch (ex)
        {
          // ignore OOM etc.
          Components.utils.reportError(ex + ' when searching on server ' + currentServer.prettyName);
          messageHeader = null;
        }
      }
    }
  }
  window.setCursor("auto");
  return messageHeader;
}


// Should messages be opened in a new window?
function CheckOpenInNewWindow()
{
  // always open new windows when called from view-source
  if (document.documentElement.id == "main-window")
    return true;
  return goMnenhy.GetPref("mailnews.messageid.openInNewWindow", goMnenhy.kiPrefBool);
}


// Overlay this function to get between the finding and the opening of a message,
// so that we can determine whether to continue the search online or not...
function OpenMessageForMessageId(asMessageId)
{
  // we need a msgWindow here!
  let oThis = CheckThisMsgWindow();
  if (!oThis)
    return;

  let oMsgHdr = FindMsgHdrByMID(asMessageId);
  if (oMsgHdr)
  {
    let bNewWindow = CheckOpenInNewWindow();
    oThis.OpenMessageByHeader(oMsgHdr, bNewWindow);
    return;
  }

  // Nothing found locally, hence search online?
  if (goMnenhy.GetPref("extensions.mnenhy.mids.online", goMnenhy.kiPrefBool))
  {
    oThis.OpenBrowserWithMessageId(asMessageId);
    return;
  }

  // (Still) Nothing found - show error message.
  let oMenu = document.getElementById("mnenhyMIDContextMenu");
  let sErrorTitle   = oMenu.getAttribute("errortitle");
  let sErrorMessage = oMenu.getAttribute("errormessage").replace(/%s/, "<" + asMessageId + ">");
  goMnenhy.AlertService(sErrorMessage, sErrorTitle, window);
}


function SearchOnNewsServer(asMessageID, asServer, aiPort)
{
  // we need a msgWindow here!
  let oThis = CheckThisMsgWindow();
  if (!oThis)
    return;

  let sURL = "news://" + asServer + ':' + aiPort +'/'+ encodeURIComponent(asMessageID);
  let bNewWindow = CheckOpenInNewWindow();
  if (bNewWindow)
  {
    if (goMnenhy.installed.IsThunderbird())
    {
      // TB3 passes URLs to an external program
      try
      {
        oThis.messenger.launchExternalURL(sURL);
      }
      catch(e)
      {
        Components.utils.reportError("Failed to open message-id in browser; "
                                   + "sURL=" + sURL);
      }
      return;
    }

    // The following code mimics utilityOverlay.js:openNewTabWindowOrExistingWith,
    // stripped down to our needs!
    
    // open in a new window if no existing window can be found
    let iLoadType = goMnenhy.GetPref("browser.link.open_external", goMnenhy.kiPrefInt);
    let oBrowserWin = (iLoadType != oThis.kNewWindow) ? oThis.getTopWin() : null;
    if (!oBrowserWin)
    {
      window.openDialog(oThis.getBrowserURL(),
                        "_blank",
                        "chrome,all,dialog=no",
                        sURL,
                        "charset=UTF-8",
                        null);
      return;
    }

    // get the existing browser object
    let oBrowser = oBrowserWin.getBrowser();
    if (iLoadType == oThis.kExistingWindow)
    {
      // change the current window/tab
      oBrowser.loadURI(sURL);
      oBrowser.docShell.QueryInterface(Components.interfaces.nsIDocCharset).charset = "UTF-8";
      oBrowserWin.content.focus();
      return;
    }

    // open in new tab
    let bLoadInBackground = goMnenhy.GetPref("browser.tabs.loadDivertedInBackground", goMnenhy.kiPrefBool);
    oBrowser.addTab(sURL, null, "UTF-8", !bLoadInBackground);
    oBrowser.docShell.QueryInterface(Components.interfaces.nsIDocCharset).charset = "UTF-8";
    if (!bLoadInBackground)
      oBrowserWin.content.focus();
  }
  else if (document.documentElement.id == "messengerWindow")
  {
    HideHeaderViews();
    // it seems that loading a message this way will always get it as UTF-8,
    // so we need to override any default charset forced by the user
    // (TB3 has no getMessageBrowser(), hence using the messagepane directly)
    document.getElementById("messagepane").docShell
            .QueryInterface(Components.interfaces.nsIDocCharset).charset = "UTF-8";
    window.content.location = sURL;
  }
}


// add all known newsservers to the context menu
function CreateNewsServerList(aoContextMenuNode)
{
  // clear out the old server list
  let roSeparators = aoContextMenuNode.getElementsByAttribute("mnenhyseptype", "mnenhyMIDServerSeparator");
  if (!roSeparators || !roSeparators.length)
    return; // huh?!
  let oShowSeparator = roSeparators[0];
  while (oShowSeparator.nextSibling)
    oShowSeparator.parentNode.removeChild(oShowSeparator.nextSibling);

  // create new server list
  let bShowSeparator = false;
  let sItemLabel = oShowSeparator.getAttribute("label");
  let sAccounts = goMnenhy.GetPref("mail.accountmanager.accounts", goMnenhy.kiPrefString);
  let rsAccounts = sAccounts.split(",");
  for (let i = 0; i < rsAccounts.length; ++i)
  {
    let sAccount    = rsAccounts[i];
    let sServer     = goMnenhy.GetPref("mail.account." + sAccount + ".server", goMnenhy.kiPrefString);
    let sServerType = goMnenhy.GetPref("mail.server."  + sServer  + ".type",   goMnenhy.kiPrefString);
    if (sServerType == "nntp")
    {
      bShowSeparator  = true;
      let sHostName   = goMnenhy.GetPref("mail.server." + sServer + ".hostname", goMnenhy.kiPrefString);
      let sServerName = goMnenhy.GetPref("mail.server." + sServer + ".name",     goMnenhy.kiPrefWString) || sHostName;
      let iServerPort = goMnenhy.GetPref("mail.server." + sServer + ".port",     goMnenhy.kiPrefInt)     || 119;
      let oItem = document.createElement("menuitem");
      let sLabel = sItemLabel.replace(/%s/, sServerName) || sServerName;
      oItem.setAttribute("label", sLabel);
      oItem.setAttribute("value", sServer);
      // pass the MID (without the surrounding <>) to the actual search function
      oItem.setAttribute("oncommand", "SearchMIDContext('" + sServer + "', event);");
      aoContextMenuNode.appendChild(oItem);
    }
  }

  // do we need to show the separator?
  oShowSeparator.setAttribute("hidden", !bShowSeparator);
}


// show the message id in the context menu
function FillMessageIdContextMenu(aoMessageIdNode)
{
  if (!aoMessageIdNode)
    return;
  let bNoMID = (aoMessageIdNode.localName == "treechildren");
  if (goMnenhy.installed.IsThunderbird())
    aoMessageIdNode = aoMessageIdNode.parentNode.parentNode;

  // if we don't have a MID, disable everything but the manual search and exit
  let sMID = aoMessageIdNode.getAttribute("messageid");
  bNoMID = bNoMID && !sMID;
  document.getElementById("mnenhyMIDLocalContext"  ).disabled = bNoMID;
  document.getElementById("mnenhyMIDBrowserContext").disabled = bNoMID;
  document.getElementById("mnenhyMIDCopyContext"   ).disabled = bNoMID;
  document.getElementById("mnenhyMIDServerSeparatorContext").hidden = bNoMID;
  if (bNoMID)
    return;

  // show the message-id in the first menuitem
  let oTitleItem = document.getElementById("messageIdContext-messageIdTarget");
  if (oTitleItem)
    oTitleItem.setAttribute("label", sMID);

  // Add the list of currently known newsservers.
  // We expect to be called from a <menupopup> or <popup> element!
  // Bug 552560 removed the __parent__ property, but provided a replacement function.
  let oNode = null;
  let oDWU = window.getInterface(Components.interfaces.nsIDOMWindowUtils);
  if (oDWU && ("getParent" in oDWU))
    oNode = oDWU.getParent(FillMessageIdContextMenu.caller);
  else
    oNode = FillMessageIdContextMenu.caller.__parent__;
  CreateNewsServerList(oNode);
}


function SearchMID(asHost, asMessageID)
{
  // no MID => nothing to do
  if (!asMessageID)
    return;

  // we need a msgWindow here!
  let oThis = CheckThisMsgWindow();
  if (!oThis)
    return;

  // search locally?
  if (!asHost || asHost == "local")
  {
    OpenMessageForMessageId(asMessageID);
    return;
  }

  // search online (e.g. Google)
  if (asHost == "browser")
  {
    oThis.OpenBrowserWithMessageId(asMessageID);
    return;
  }

  // just copy the string
  if (asHost == "copy")
  {
    oThis.CopyString("<" + asMessageID + ">");
    return;
  }

  // we now assume a real server, hence get its port and do the search
  let sHostName   = goMnenhy.GetPref("mail.server." + asHost + ".hostname", goMnenhy.kiPrefString);
  let iServerPort = goMnenhy.GetPref("mail.server." + asHost + ".port",     goMnenhy.kiPrefInt) || 119;
  SearchOnNewsServer(asMessageID, sHostName, iServerPort);
}


function SearchMIDManual()
{
  // open the input dialog for message
  let oReturnValue = {msMID: "", msLocation: ""};
  window.openDialog("chrome://mnenhy-mids/content/mnenhy-mids-input.xul",
                    "mnenhyMIDSDialog",
                    "center,modal",
                    oReturnValue);
  // do nothing if no MID was returned
  if (!oReturnValue.msMID)
    return;
  SearchMID(oReturnValue.msLocation, oReturnValue.msMID);
}


// get the aiN-th parent node of aoNode, if it exists, or null
// aiN must be >= 0
function GetNthParent(aoNode, aiN)
{
  if (aiN < 0)
    return null;
  while ((--aiN >= 0) && aoNode && aoNode.parentNode)
    aoNode = aoNode.parentNode;
  return aoNode;
}


function SearchMIDContext(asLocation, aoEvent)
{
  let sMessageID = "";

  // first determine where are we called from
  let oTarget = GetNthParent(aoEvent.explicitOriginalTarget, 1);
  if (oTarget && oTarget.id == "messageIdContext")
  {
    // message-id headerpane context
    sMessageID = GetMessageIdFromNode(document.popupNode, true);
    SearchMID(asLocation, sMessageID);
    return;
  }

  // if the popupnode is an anchor, we try to parse the link (a bit)
  // (nodeName/tagName may contain namespaces, the case of localName for HTML
  //  changed between m-c 1.9.1 and 1.9.2)
  let sNodeName = document.popupNode.localName.toLowerCase();
  if (sNodeName == "a")
  {
    // get the link, delete a leading protocol and any trailing slashes, add a trailing /
    // (we may not have a gContextMenu here, so we can't use getLinkURL etc.)
    sMessageID = "/" + document.popupNode.href.replace(/^.*?:\/*|\/+$/g, "");
    // now assume that anything after the last / is the MID and decode it
    sMessageID = decodeURIComponent(sMessageID.replace(/^.*\//g, ""));
    SearchMID(asLocation, sMessageID);
    return;
  }

  // In all other cases, we expect to find a selection (somewhere)!
  oTarget = GetNthParent(aoEvent.explicitOriginalTarget, 3);
  if (oTarget && oTarget.id == "MnenhyTextPopup")
  {
    // we're called from a message header context menu
    // first determine the actual header element
    let oHeaderNode = document.popupNode;
    while (oHeaderNode && (oHeaderNode.getAttribute("context") != "MnenhyTextPopup"))
      oHeaderNode = oHeaderNode.parentNode;
    if (oHeaderNode)
    {
      // reached the base node of this header value
      // grab the selection...
      let oSelection = oHeaderNode.ownerDocument.defaultView.getSelection();
      // ... but only allow those selections which are fully part of this header
      let oAnchorNode = oSelection.anchorNode;
      while (oAnchorNode && oAnchorNode != oHeaderNode)
        oAnchorNode = oAnchorNode.parentNode;
      if (!oAnchorNode)
        oSelection.removeAllRanges();  // invalidate "external" selections
      // read the selected text, which shall represent a message-id
      sMessageID = oSelection.toString();
      // if nothing was selected, use the full header value
      if (!sMessageID)
        sMessageID = oHeaderNode.textContent;
    }
  }
  else
  {
    // not a message header, thus use the window's content selection
    sMessageID = window.content.getSelection().toString();
  }

  // search for the selection without surrounding spaces and angles
  SearchMID(asLocation, sMessageID.replace(/^[ <]+|[ >]+$/g, ""));
}
