/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mnenhy.
 *
 * The Initial Developer of the Original Code is Karsten Düsterloh <kd@mnenhy.de>.
 *
 * Portions created by the Initial Developer are Copyright © 2002-2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// +-
// |  the contents of this file will be loaded into the Mnenhy.installed.* context!
// |  ie. foo() is available as goMnenhy.installed.foo()
// +-

// |
// |  constants
// |

// debug
const gksInstalledDebugPref  = "extensions.mnenhy.log.installed";
// Mnenhy module status pref
// We use negative values, so that unset prefs will mean active packages.
const gksPackageDisabledPref = "extensions.mnenhy.disabled.";

// XUL.mfasl
const gksCID_Fastload        = "@mozilla.org/fast-load-service;1";
const gksMozillaXUL_Fastload = "XUL";

// Mozilla applications (bit flags)
// - base bit flags
const kiMozillaBrandIDUnknown     = 0;
const kiMozillaBrandIDMozilla     = 1 << 0;   // includes SeaMonkey
const kiMozillaBrandIDThunderbird = 1 << 1;
const kiMozillaBrandIDFirefox     = 1 << 2;
// - composite flags
const kiMozillaBrandIDMail        = kiMozillaBrandIDMozilla | kiMozillaBrandIDThunderbird;
const kiMozillaBrandIDAll         = kiMozillaBrandIDMail    | kiMozillaBrandIDFirefox;

// Information about all existing Mnenhy subpackages:
const kiIndexAllowedApp  = 0; // bit flags: which applications use this package?
const kiIndexHasInitFile = 1; // bool: has an init file
const kiIndexInitContext = 2; // string: context to load init file into
const kiIndexInitFile    = 3; // string: init file URI
const koMnenhyPackages   =
    {"mnenhy-headers":    [kiMozillaBrandIDMail,    true, "headers",     "chrome://mnenhy-headers/content/mnenhy-headers.js"],
     "mnenhy-folderstore":[kiMozillaBrandIDMail,    true, "folderstore", "chrome://mnenhy-folderstore/content/mnenhy-folderstore.js"],
     "mnenhy-codecs":     [kiMozillaBrandIDAll,     true, "codecs",      "chrome://mnenhy-codecs/content/mnenhy-codecs.js"],
     "mnenhy-mids":       [kiMozillaBrandIDMail,    false],
     "mnenhy-junk":       [kiMozillaBrandIDMail,    false],
     "mnenhy-sidebar":    [kiMozillaBrandIDMozilla, false],
     "mnenhy":            [kiMozillaBrandIDAll,     false] };
const ksInstalledPrefix = "chrome://";
// Indices for GetAllPackages
const kiIndexInfoName     = 0; // string: package name
const kiIndexInfoAllowed  = 1; // bool: package allowed?
const kiIndexInfoInstall  = 2; // bool: package installed?
const kiIndexInfoActive   = 3; // bool: package active?


// |
// |  global variables
// |
let giMozillaApp         = kiMozillaBrandIDMozilla; // which Mozilla application is running?
let goMnenhyManifestFile = null;  // nsIFile
let gbInstalledDebug     = null;  // dump debug info

// |
// | functions
// |

// remove $UserProfileDir/xul.mfl (resp. xul.mfasl)
function DeleteXulFastload()
{
  debugdump("installed.DeleteXulFastload+");
  try
  {
    let oFaslService = Components.classes[gksCID_Fastload]
                                 .getService(Components.interfaces.nsIFastLoadService);
    let oFaslFile    = oFaslService.newFastLoadFile(gksMozillaXUL_Fastload);
    oFaslFile.remove(false);
  }
  catch(ignored){}
  debugdump("installed.DeleteXulFastload-");
}


// Get the chrome.manifest file for Mnenhy
// Return: an nsIFile
function GetMnenhyManifestFile()
{
  if (!goMnenhyManifestFile)
  {
    // We assume the following here:
    // - chrome.manifest lies in directory x
    // - Mnenhy's base package directory is x/chrome
    // That way, we can grab the local path to the base package and cut it down to x!
    let rsMnenhyManifest = goMnenhy.ConvertURI2FileURI("chrome://mnenhy/content/").spec.match(/(.*\/chrome)\/.*$/);
    debugdump("installed.GetMnenhyManifestFile: rsMnenhyManifest=" + rsMnenhyManifest);
    if (rsMnenhyManifest && rsMnenhyManifest.length > 1)
    {
      // \1 should now be something like "file:///path/to/Mnenhy-GUID/chrome"
      goMnenhyManifestFile = goMnenhy.ConvertURI2FileURI(rsMnenhyManifest[1] + ".manifest").file;
    }
    debugdump("installed.GetMnenhyManifestFile=" + goMnenhyManifestFile);
  }
  return goMnenhyManifestFile;
}


// SeaMonkey had a dedicated function for temporarily deactivating certain packages -
// this has been dropped in the new toolkit registry.
// We simulate this by keeping that package in the manifest file, but commenting it out.
// We also store the package activation states in the preferences,
// as extensions.mnenhy.disabled.$asPackage = !$abAllowed.
function SetAllowOverlaysForPackage(asPackage, abAllowed)
{
  debugdump("installed.SetAllowOverlaysForPackage(" + asPackage + "," + abAllowed +")");
  if (!asPackage)
    return;

  // set pref
  goMnenhy.SetPref(gksPackageDisabledPref + asPackage, goMnenhy.kiPrefBool, !abAllowed);

  // now manipulate the manifest, if necessary
  let oMnenhyManifestFile = GetMnenhyManifestFile();
  if (oMnenhyManifestFile)
  {
    // we match lines as "everything interesting between to LF",
    // hence we need to make sure this will always work
    let sMnenhyManifest = "\n" + goMnenhy.ReadFile(oMnenhyManifestFile) + "\n";

    // does the manifest file contain our package?
    let sPackageLine = "\\n\\s*#?\\s*((content|locale|skin)\\s+" + asPackage + "\\s.*)";
    let xPackageLine = new RegExp(sPackageLine);
    // the global search gets all matching lines, but no brace-matches
    let rsPackageLines = sMnenhyManifest.match(new RegExp(sPackageLine, "g"));
    debugdump("installed.SetAllowOverlaysForPackage: rsPackageLines=" + rsPackageLines);
    for (let i in rsPackageLines)
    {
      // (don't) comment these lines out (if match[1] wouldn't exist, we weren't here in the first place...)
      let sNewPackageLine = "\n" + (abAllowed ? "" : "#") + rsPackageLines[i].match(xPackageLine)[1];
      sMnenhyManifest = sMnenhyManifest.replace(rsPackageLines[i], sNewPackageLine);
    }

    // we need to touch all overlays and overrides of this package as well
    let sOverLine = "\\n\\s*#?\\s*((overlay|override)\\s+.*?\\s+chrome://" + asPackage + "/.*)";
    let xOverLine = new RegExp(sOverLine);
    // the global search gets all matching lines, but no brace-matches
    let rsOverLines = sMnenhyManifest.match(new RegExp(sOverLine, "g"));
    debugdump("installed.SetAllowOverlaysForPackage: rsOverKnown=" + rsOverLines);
    for (let j in rsOverLines)
    {
      // (don't) comment these lines out (if match[1] wouldn't exist, we weren't here in the first place...)
      let sNewOverLine = "\n" + (abAllowed ? "" : "#") + rsOverLines[j].match(xOverLine)[1];
      sMnenhyManifest = sMnenhyManifest.replace(rsOverLines[j], sNewOverLine);
    }

    // and write the manifest back if it changed
    if (rsPackageLines || rsOverLines)
    {
      try
      {
        // remove surrounding whitespace
        sMnenhyManifest = sMnenhyManifest.replace(/^\s+|\s+$/g, "");
        goMnenhy.WriteFile(oMnenhyManifestFile, sMnenhyManifest);
      }
      catch (ignored) 
      {
        // couldn't change the manifest, most probably because we lack write permissions
        debugdump("installed.SetAllowOverlaysForPackage: couldn't update Mnenhy's chrome.manifest!\n" +
                  "Insufficient write permissions for '" + oMnenhyManifestFile.spec + "'?");
      }
    }
  }
}


// return an array of arrays ["package name" (, bIsAllowed, bIsInstalled, bIsActive)]
// data in () is only returned if abShowActive is true
function GetAllPackages(abShowActive)
{
  let rRet = [], rh, oh;
  for (let s in koMnenhyPackages)
  {
    rh = [];
    rh.push(s); // package name
    if (abShowActive)
    {
      oh = GetPackageStatus(s);
      rh.push(oh.bIsAllowed);   // package allowed?
      rh.push(oh.bIsInstalled); // package installed?
      rh.push(oh.bIsActive);    // package active?
    }
    rRet.push(rh);
  }
  return rRet;
}


// Is this Mnenhy package installed, i.e. registered with chrome? Is it active?
//  IN:     string  asPackage Mnenhy package name
//  RETURN: object  {bool bIsAllowed, bool bIsInstalled, bool bIsActive}
function GetPackageStatus(asPackage)
{
  let oRet = {bIsAllowed: false, bIsInstalled: false, bIsActive: false};
  if (asPackage in koMnenhyPackages)
  {
    // grab permission from config array
    oRet.bIsAllowed = !!(koMnenhyPackages[asPackage][kiIndexAllowedApp] & giMozillaApp);
    // read activation state from Mnenhy's chrome.manifest:
    // - installed packages appear there as a "content" line
    // - inactive packages appear there as comment-prefixed (#) "content" line
    let oMnenhyManifestFile = GetMnenhyManifestFile();
    if (oMnenhyManifestFile)
    {
      // we match lines as "everything interesting between to LF",
      // hence we need to make sure this will always work
      let sMnenhyManifest = "\n" + goMnenhy.ReadFile(oMnenhyManifestFile) + "\n";
      // does the manifest file contain our package?
      let rsPackageKnown = sMnenhyManifest.match(new RegExp("\\n\\s*(#?)\\s*content\\s+" + asPackage + "\\s"));
      debugdump("installed.GetPackageStatus: rsPackageKnown=" + rsPackageKnown);
      oRet.bIsInstalled = !!rsPackageKnown;  // have package
      oRet.bIsActive    = !!rsPackageKnown && (rsPackageKnown.length > 1) && (rsPackageKnown[1] == ""); // not a comment
    }
  }
  debugdump("installed.GetPackageStatus for "+asPackage+" = "+oRet.toSource());
  return oRet;
}


function IsMozilla()
{
  return giMozillaApp == kiMozillaBrandIDMozilla;
}
function IsThunderbird()
{
  return giMozillaApp == kiMozillaBrandIDThunderbird;
}
function IsFirefox()
{
  return giMozillaApp == kiMozillaBrandIDFirefox;
}


// debug
function debugdump(aoDump)
{
  if (gbInstalledDebug)
    goMnenhy.Log(aoDump);
}


// |
// |  Init called by the global Mnenhy object when loading this script file
// |
function Init(aoMnenhy)
{
  // keep pointer to have the Mnenhy object prototype available
  this.goMnenhy    = aoMnenhy;
  gbInstalledDebug = goMnenhy.GetPref(gksInstalledDebugPref, goMnenhy.kiPrefBool);

  // get the currently running application type
  let oXULAppInfo = Components.classes["@mozilla.org/xre/app-info;1"]
                              .getService(Components.interfaces.nsIXULAppInfo);
  switch (oXULAppInfo.ID)
  {
    case "{92650c4d-4b8e-4d2a-b7eb-24ecf4f6b63a}":
      giMozillaApp = kiMozillaBrandIDMozilla;
      break;
    case "{ec8030f7-c20a-464f-9b0e-13a3a9e97384}":
      giMozillaApp = kiMozillaBrandIDFirefox;
      break;
    case "{3550f703-e582-4d05-9a08-453d09bdfdc6}":
      giMozillaApp = kiMozillaBrandIDThunderbird;
      break;
    default:
      giMozillaApp = kiMozillaBrandIDUnknown;
  }
  debugdump("installed.Init: giMozillaApp="+giMozillaApp);

  // load and init installed packages (if necessary)
  let bRestartNeeded = false;
  let rh, sh, oh, rsIllegalPackages = [];
  for (let s in koMnenhyPackages)
  {
    rh = koMnenhyPackages[s];
    // only check installed packages
    oh = GetPackageStatus(s);
    if (oh.bIsInstalled)
    {
      if (oh.bIsAllowed)
      {
        // The activation state in the preferences extensions.mnenhy.disabled.*
        // has priority over the manifest status! That way, a package's disabled
        // status will survive extension updates (which write new manifests!).
        let bEnabledByPref = !goMnenhy.GetPref(gksPackageDisabledPref + s, goMnenhy.kiPrefBool);
        if (bEnabledByPref != oh.bIsActive)
        {
          // if pref and manifest disagree, pref wins
          SetAllowOverlaysForPackage(s, bEnabledByPref);
          // we didn't require a restart yet, check whether we changed the manifest
          // and require a restart then (if we'd require a restart without really
          // changing the activation state, we'd get a restart loop!)
          if (!bRestartNeeded)
          {
            let oCheck = GetPackageStatus(s);
            if (bEnabledByPref == oCheck.bIsActive)
              bRestartNeeded = true;
          }
        }
        // init only allowed, active packages
        if (bEnabledByPref && rh[kiIndexHasInitFile])
        {
          // need to load and init
          sh = rh[kiIndexInitContext];  // get pseudo namespace name
          goMnenhy[sh] = {};            // create pseudo namespace
          if (goMnenhy.Include(rh[kiIndexInitFile], goMnenhy[sh], true))
            goMnenhy[sh].Init(aoMnenhy);
        } // need init
      }
      else
      {
        // Not allowed! If active, deactivate it.
        if (oh.bIsActive)
        {
          goMnenhy.Log("Package '" + s + "' not allowed for "+goMnenhy.ksMozillaBrandName+"!")
          SetAllowOverlaysForPackage(s, false);
          rsIllegalPackages.push(s);
        }
      }
    } // is installed
  } // for package init
  debugdump("installed.Init:: illegal packages="+rsIllegalPackages)

  // If we found illegal active packages, we need to invalidate the XUL.mfasl
  // to make the package deactivation happen on next restart!
  if (rsIllegalPackages.length)
    DeleteXulFastload();

  // Finally, if we made startup(!) changes to the manifest file, we need to reboot!
  if (bRestartNeeded)
  {
    const kCiAS = Components.interfaces.nsIAppStartup;
    Components.classes['@mozilla.org/toolkit/app-startup;1']
              .getService(kCiAS)
              .quit(kCiAS.eForceQuit | kCiAS.eRestart); // force restart
  }
}
